use bytemuck;
use galactica_content::Content;
use wgpu;
use winit;

use crate::{starfield::Starfield, texturearray::TextureArray, ui::UiManager, RenderState};

/// GPUState is very big, so its methods have been split
/// among the following files.
mod new;
mod phys;
mod render;

/// A high-level GPU wrapper. Reads game state (via RenderInput),
/// produces pretty pictures.
pub struct GPUState {
	device: wgpu::Device,
	config: wgpu::SurfaceConfiguration,
	surface: wgpu::Surface,

	object_pipeline: wgpu::RenderPipeline,
	starfield_pipeline: wgpu::RenderPipeline,
	particle_pipeline: wgpu::RenderPipeline,
	ui_pipeline: wgpu::RenderPipeline,
	radialbar_pipeline: wgpu::RenderPipeline,

	starfield: Starfield,
	texture_array: TextureArray,

	state: RenderState,
	ui: UiManager,
}

impl GPUState {
	/// Get the window we are attached to
	pub fn window(&self) -> &winit::window::Window {
		&self.state.window
	}

	/// Update window size.
	/// This should be called whenever our window is resized.
	pub fn resize(&mut self, ct: &Content) {
		let new_size = self.state.window.inner_size();
		if new_size.width > 0 && new_size.height > 0 {
			self.state.window_size = new_size;
			self.state.window_aspect = new_size.width as f32 / new_size.height as f32;
			self.config.width = new_size.width;
			self.config.height = new_size.height;
			self.surface.configure(&self.device, &self.config);
		}
		self.starfield.update_buffer(ct, &mut self.state);
	}

	/// Initialize the rendering engine
	pub fn init(&mut self, ct: &Content) {
		// Update global values
		self.state.queue.write_buffer(
			&self.state.global_uniform.atlas_buffer,
			0,
			bytemuck::cast_slice(&[self.texture_array.image_locations]),
		);
		self.state.queue.write_buffer(
			&self.state.global_uniform.sprite_buffer,
			0,
			bytemuck::cast_slice(&[self.texture_array.sprite_data]),
		);

		self.starfield.update_buffer(ct, &mut self.state);
	}
}
