use galactica_content::Relationship;
use galactica_util::clockwise_angle;
use nalgebra::Vector2;
use rapier2d::{dynamics::RigidBodySet, geometry::ColliderHandle};
use std::collections::HashMap;

use crate::data::ShipState;

use super::{
	super::{
		objects::{PhysSimShip, ShipControls},
		PhysStepResources,
	},
	ShipControllerStruct,
};

/// "Point" ship controller.
/// Point and shoot towards the nearest enemy.
#[derive(Debug, Clone)]
pub struct PointShipController {}

impl PointShipController {
	/// Create a new ship controller
	pub fn new() -> Self {
		Self {}
	}
}

impl ShipControllerStruct for PointShipController {
	fn update_controls(
		&mut self,
		res: &PhysStepResources,
		rigid_bodies: &RigidBodySet,
		ships: &HashMap<ColliderHandle, PhysSimShip>,
		this_ship: ColliderHandle,
	) -> Option<ShipControls> {
		let mut controls = ShipControls::new();

		let my_ship = ships.get(&this_ship).unwrap();
		let this_rigidbody = rigid_bodies.get(my_ship.rigid_body).unwrap();
		let my_position = this_rigidbody.translation();
		let my_rotation = this_rigidbody.rotation();
		let my_angvel = this_rigidbody.angvel();
		let my_faction = res.ct.get_faction(my_ship.data.get_faction());

		// Iterate all possible targets
		let mut hostile_ships = ships
			.values()
			.filter(
				// Target only flying ships we're hostile towards
				|s| match my_faction.relationships.get(&s.data.get_faction()).unwrap() {
					Relationship::Hostile => match s.data.get_state() {
						ShipState::Flying { .. } => true,
						_ => false,
					},
					_ => false,
				},
			)
			.map(|s| rigid_bodies.get(s.rigid_body).unwrap());

		// Find the closest target
		let mut closest_enemy_position = match hostile_ships.next() {
			Some(c) => c.translation(),
			None => return Some(controls), // Do nothing if no targets are available
		};

		let mut d = (my_position - closest_enemy_position).magnitude();
		for r in hostile_ships {
			let p = r.translation();
			let new_d = (my_position - p).magnitude();
			if new_d < d {
				d = new_d;
				closest_enemy_position = p;
			}
		}

		let angle = clockwise_angle(
			&(my_rotation * Vector2::new(1.0, 0.0)),
			&(closest_enemy_position - my_position),
		);

		if angle < 0.0 && my_angvel > -0.3 {
			controls.right = true;
		} else if angle > 0.0 && my_angvel < 0.3 {
			controls.left = true;
		}

		controls.guns = true;
		return Some(controls);
	}
}
