use galactica_content::{CollapseEvent, Ship};
use nalgebra::{Point2, Vector2};
use rand::{rngs::ThreadRng, Rng};
use rapier2d::{dynamics::RigidBody, geometry::Collider};

use super::super::{ParticleBuilder, PhysStepResources};
use crate::data::ShipData;

#[derive(Debug, Clone)]
pub(super) struct ShipCollapseSequence {
	rng: ThreadRng,

	/// The total length of this collapse sequence
	total_length: f32,

	/// How many seconds we've spent playing this sequence
	elapsed: f32,
}

impl ShipCollapseSequence {
	pub(super) fn new(total_length: f32) -> Self {
		Self {
			rng: rand::thread_rng(),
			total_length,
			elapsed: 0.0,
		}
	}

	/// Has this collapse sequence fully played out?
	pub fn is_done(&self) -> bool {
		self.elapsed >= self.total_length
	}

	/// Pick a random points inside a ship's collider
	fn random_in_ship(&mut self, ship: &Ship, collider: &Collider) -> Vector2<f32> {
		let mut y = 0.0;
		let mut x = 0.0;
		let mut a = false;
		while !a {
			x = self.rng.gen_range(-1.0..=1.0) * ship.size / 2.0;
			y = self.rng.gen_range(-1.0..=1.0) * ship.size * ship.sprite.aspect / 2.0;
			a = collider.shape().contains_local_point(&Point2::new(x, y));
		}
		Vector2::new(x, y)
	}

	/// Step this sequence `t` seconds
	pub(super) fn step(
		&mut self,
		res: &mut PhysStepResources,
		ship_data: &ShipData,
		rigid_body: &mut RigidBody,
		collider: &mut Collider,
	) {
		let ship_content = res.ct.get_ship(ship_data.get_content());
		let ship_pos = rigid_body.translation();
		let ship_rot = rigid_body.rotation();

		// The fraction of this collapse sequence that has been played
		let frac_done = self.elapsed / self.total_length;

		// TODO: slight random offset for event particles

		// Trigger collapse events
		for event in &ship_content.collapse.events {
			match event {
				CollapseEvent::Effect(event) => {
					if (event.time > self.elapsed && event.time <= self.elapsed + res.t)
						|| (event.time == 0.0 && self.elapsed == 0.0)
					// ^^ Don't miss events scheduled at the very start of the sequence!
					{
						for spawner in &event.effects {
							let effect = res.ct.get_effect(spawner.effect);

							for _ in 0..spawner.count as usize {
								let pos: Vector2<f32> = if let Some(pos) = spawner.pos {
									Vector2::new(pos.x, pos.y)
								} else {
									self.random_in_ship(ship_content, collider)
								};
								let pos = ship_pos + (ship_rot * pos);

								let velocity =
									rigid_body.velocity_at_point(&Point2::new(pos.x, pos.y));

								res.particles.push(ParticleBuilder::from_content(
									effect,
									pos.into(),
									0.0,
									velocity,
									Vector2::new(0.0, 0.0),
								))
							}
						}
					}
				}
			}
		}

		// Create collapse effects
		for spawner in &ship_content.collapse.effects {
			let effect = res.ct.get_effect(spawner.effect);

			// Probability of adding a particle this frame.
			// The area of this function over [0, 1] should be 1.
			let pdf = |x: f32| {
				let f = 0.2;
				let y = if x < (1.0 - f) {
					let x = x / (1.0 - f);
					(x * x + 0.2) * 1.8 - f
				} else {
					1.0
				};
				return y;
			};

			let p_add = (res.t / self.total_length) * pdf(frac_done) * spawner.count;

			if self.rng.gen_range(0.0..=1.0) <= p_add {
				let pos = if let Some(pos) = spawner.pos {
					Vector2::new(pos.x, pos.y)
				} else {
					self.random_in_ship(ship_content, collider)
				};

				// Position, adjusted for ship rotation
				let pos = ship_pos + (ship_rot * pos);
				let vel = rigid_body.velocity_at_point(&Point2::new(pos.x, pos.y));

				res.particles.push(ParticleBuilder {
					sprite: effect.sprite,
					pos: pos.into(),
					velocity: vel,
					angle: 0.0,
					angvel: 0.0,
					lifetime: effect.lifetime,
					size: effect.size,
					fade: 0.0,
				});
			}
		}

		self.elapsed += res.t;
	}
}
