use galactica_content::{Effect, SpriteHandle};
use nalgebra::{Point2, Rotation2, Vector2};
use rand::Rng;

/// Instructions to create a new particle
#[derive(Debug)]
pub struct ParticleBuilder {
	/// The sprite to use for this particle
	pub sprite: SpriteHandle,

	/// This object's center, in world coordinates.
	pub pos: Point2<f32>,

	/// This particle's velocity, in world coordinates
	pub velocity: Vector2<f32>,

	/// This particle's angle, in radians
	pub angle: f32,

	/// This particle's angular velocity (rad/sec)
	pub angvel: f32,

	/// This particle's lifetime, in seconds
	pub lifetime: f32,

	/// The size of this particle,
	/// given as height in world units.
	pub size: f32,

	/// Fade this particle out over this many seconds as it expires
	pub fade: f32,
}

impl ParticleBuilder {
	/// Create a ParticleBuilder from an Effect
	pub fn from_content(
		effect: &Effect,
		pos: Point2<f32>,
		parent_angle: f32,
		parent_velocity: Vector2<f32>,
		target_velocity: Vector2<f32>,
	) -> Self {
		let mut rng = rand::thread_rng();

		let velocity = {
			let a =
				rng.gen_range(-effect.velocity_scale_parent_rng..=effect.velocity_scale_parent_rng);
			let b =
				rng.gen_range(-effect.velocity_scale_target_rng..=effect.velocity_scale_target_rng);

			let velocity = ((effect.velocity_scale_parent + a) * parent_velocity)
				+ ((effect.velocity_scale_target + b) * target_velocity);

			Rotation2::new(rng.gen_range(-effect.direction_rng..=effect.direction_rng)) * velocity
		};

		let angvel = if effect.angvel_rng == 0.0 {
			effect.angvel
		} else {
			effect.angvel + rng.gen_range(-effect.angvel_rng..=effect.angvel_rng)
		};
		let angle = if effect.angle_rng == 0.0 {
			parent_angle + effect.angle
		} else {
			parent_angle + effect.angle + rng.gen_range(-effect.angle_rng..=effect.angle_rng)
		};

		ParticleBuilder {
			sprite: effect.sprite,
			pos,
			velocity,

			angle,
			angvel,

			lifetime: 0f32
				.max(effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng)),

			// Make sure size isn't negative. This check should be on EVERY rng!
			size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),

			fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
		}
	}
}
