use galactica_content::{Content, FactionHandle, ShipHandle, SystemHandle};
use galactica_playeragent::PlayerAgent;
use nalgebra::{Point2, Vector2};

use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle, RigidBodySet},
	geometry::{ColliderHandle, ColliderSet, Group, InteractionGroups},
};
use std::collections::HashMap;

use crate::data::{ShipAutoPilot, ShipPersonality, ShipState};

use super::{
	controller::ShipController,
	objects::{PhysProjectile, PhysSimShip},
};

// TODO: replace with a more generic handle
/// A handle for a ship in this simulation
/// This lets other crates reference ships
/// without including `rapier2d`.
pub struct PhysSimShipHandle(pub ColliderHandle);

/// Manages the physics state of one system
#[derive(Clone)]
pub struct PhysSim {
	/// The system this sim is attached to
	_system: SystemHandle,

	rigid_body_set: RigidBodySet,
	collider_set: ColliderSet,

	projectiles: HashMap<ColliderHandle, PhysProjectile>,
	ships: HashMap<ColliderHandle, PhysSimShip>,
	ship_behaviors: HashMap<ColliderHandle, ShipController>,
}

mod step;
mod steputil;

// Public methods
impl PhysSim {
	/// Create a new physics system
	pub fn new(_ct: &Content, system: SystemHandle) -> Self {
		Self {
			_system: system,
			rigid_body_set: RigidBodySet::new(),
			collider_set: ColliderSet::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			ship_behaviors: HashMap::new(),
		}
	}

	/// Add a ship to this physics system
	pub fn add_ship(
		&mut self,
		ct: &Content,
		handle: ShipHandle,
		faction: FactionHandle,
		personality: ShipPersonality,
		position: Point2<f32>,
	) -> PhysSimShipHandle {
		let ship_content = ct.get_ship(handle);
		let mut cl = ship_content.collider.0.clone();
		// TODO: additonal ship mass from outfits and cargo

		let rb = RigidBodyBuilder::dynamic()
			.angular_damping(ship_content.angular_drag)
			.linear_damping(ship_content.linear_drag)
			.translation(Vector2::new(position.x, position.y))
			.can_sleep(false);

		cl.set_collision_groups(InteractionGroups::new(
			Group::GROUP_1,
			Group::GROUP_1 | Group::GROUP_2,
		));

		let ridid_body = self.rigid_body_set.insert(rb.build());
		let collider =
			self.collider_set
				.insert_with_parent(cl, ridid_body, &mut self.rigid_body_set);

		self.ship_behaviors.insert(
			collider,
			match personality {
				ShipPersonality::Dummy | ShipPersonality::Player => ShipController::new_null(),
				ShipPersonality::Point => ShipController::new_point(),
			},
		);

		self.ships.insert(
			collider,
			PhysSimShip::new(ct, handle, personality, faction, ridid_body, collider),
		);

		return PhysSimShipHandle(collider);
	}

	/// Update a player ship's controls
	pub fn update_player_controls(&mut self, ct: &Content, player: &PlayerAgent) {
		if player.ship.is_none() {
			return;
		}
		let ship_object = self.ships.get_mut(&player.ship.unwrap());
		if let Some(ship_object) = ship_object {
			match ship_object.data.get_state() {
				ShipState::Landing { .. }
				| ShipState::UnLanding { .. }
				| ShipState::Collapsing { .. }
				| ShipState::Dead => {}

				ShipState::Flying {
					autopilot: ShipAutoPilot::None,
				} => {
					ship_object.controls.guns = player.input.pressed_guns();
					ship_object.controls.left = player.input.pressed_left();
					ship_object.controls.right = player.input.pressed_right();
					ship_object.controls.thrust = player.input.pressed_thrust();

					if player.input.pressed_land() {
						ship_object.data.set_autopilot(ShipAutoPilot::Landing {
							target: player.selection.get_planet().unwrap(),
						})
					}
				}

				ShipState::Flying { .. } => {
					// Any input automatically releases autopilot
					if player.input.pressed_left()
						|| player.input.pressed_right()
						|| player.input.pressed_thrust()
						|| player.input.pressed_guns()
					{
						ship_object.data.set_autopilot(ShipAutoPilot::None);
					}
				}

				ShipState::Landed { .. } => {
					if player.input.pressed_land() {
						self.start_unland_ship(ct, player.ship.unwrap());
					}
				}
			};
		}
	}
}

// Public getters
impl PhysSim {
	/// Get a ship physics object
	pub fn get_ship(&self, ship: &PhysSimShipHandle) -> Option<&PhysSimShip> {
		self.ships.get(&ship.0)
	}

	/// Get a ship physics object
	pub fn get_ship_mut(&mut self, ship: &PhysSimShipHandle) -> Option<&mut PhysSimShip> {
		self.ships.get_mut(&ship.0)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> Option<&RigidBody> {
		self.rigid_body_set.get(r)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body_mut(&mut self, r: RigidBodyHandle) -> Option<&mut RigidBody> {
		self.rigid_body_set.get_mut(r)
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ship_body(&self) -> impl Iterator<Item = (&PhysSimShip, &RigidBody)> + '_ {
		self.ships
			.values()
			.map(|x| (x, self.rigid_body_set.get(x.rigid_body).unwrap()))
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ships(&self) -> impl Iterator<Item = &PhysSimShip> + '_ {
		self.ships.values()
	}

	/// Iterate over all ships in this physics system
	pub fn iter_projectiles(&self) -> impl Iterator<Item = &PhysProjectile> + '_ {
		self.projectiles.values()
	}
}
