use crossbeam::channel::Receiver;
use rapier2d::{
	dynamics::{
		CCDSolver, ImpulseJointSet, IntegrationParameters, IslandManager, MultibodyJointSet,
		RigidBodySet,
	},
	geometry::{BroadPhase, ColliderSet, CollisionEvent, NarrowPhase},
	na::vector,
	pipeline::{ChannelEventCollector, PhysicsPipeline},
};

/// Wraps Rapier2d physics parameters
pub struct Wrapper {
	pub(crate) ip: IntegrationParameters,
	pub(crate) pp: PhysicsPipeline,
	pub(crate) im: IslandManager,
	pub(crate) bp: BroadPhase,
	pub(crate) np: NarrowPhase,
	pub(crate) ij: ImpulseJointSet,
	pub(crate) mj: MultibodyJointSet,
	pub(crate) ccd: CCDSolver,

	collision_handler: ChannelEventCollector,

	/// Collision event queue
	/// this should be emptied after every frame.
	pub collision_queue: Receiver<CollisionEvent>,
}

impl Wrapper {
	/// Make a new physics wrapper
	pub fn new() -> Self {
		let (collision_send, collision_queue) = crossbeam::channel::unbounded();
		let (contact_force_send, _) = crossbeam::channel::unbounded();

		Self {
			ip: IntegrationParameters::default(),
			pp: PhysicsPipeline::new(),
			im: IslandManager::new(),
			bp: BroadPhase::new(),
			np: NarrowPhase::new(),
			ij: ImpulseJointSet::new(),
			mj: MultibodyJointSet::new(),
			ccd: CCDSolver::new(),
			collision_queue,
			collision_handler: ChannelEventCollector::new(collision_send, contact_force_send),
		}
	}

	/// Step physics sim by `t` seconds
	pub fn step(
		&mut self,
		t: f32,
		rigid_body_set: &mut RigidBodySet,
		collider_set: &mut ColliderSet,
	) {
		self.ip.dt = t;
		self.pp.step(
			&vector![0.0, 0.0],
			&self.ip,
			&mut self.im,
			&mut self.bp,
			&mut self.np,
			rigid_body_set,
			collider_set,
			&mut self.ij,
			&mut self.mj,
			&mut self.ccd,
			None,
			&(),
			&mut self.collision_handler,
		);
	}
}
